"""OTE-API OPTIMADE

A plugin for OTE-API.

Authored by Casper Welzel Andersen, SINTEF, 2022
Created from cookiecutter-oteapi-plugin, SINTEF, 2022
"""

__version__ = "0.2.2"
__author__ = "Casper Welzel Andersen"
__author_email__ = "casper.w.andersen@sintef.no"
